﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Collections;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Diagnostics;

namespace ASPItalia.Books.Chapter07.Components
{
    public class VirtualList<T> : IList<T>, INotifyCollectionChanged
    {

        private object syncObject = new object();
        private List<T> loadedList = new List<T>();
        private List<Int32> askedStartIndex = new List<Int32>();
        private T loadItem;

        public event NotifyCollectionChangedEventHandler CollectionChanged;

        public event EventHandler<ItemsNeededEventArgs> ItemsNeeded;

        public VirtualList(T loadItem)
        {
            this.loadItem = loadItem;
        }

        #region Not supported

        public void Clear()
        {
            throw new NotSupportedException();
        }

        public void Insert(int index, T value)
        {
            throw new NotSupportedException();
        }

        public void CopyTo(T[] array, int index)
        {
            throw new NotSupportedException();
        }

        public bool Remove(T value)
        {
            throw new NotSupportedException();
        }

        public void RemoveAt(int index)
        {
            throw new NotSupportedException();
        }

        #endregion

        #region Fixed implementation

        public bool IsFixedSize
        {
            get { return true; }
        }

        public bool IsReadOnly
        {
            get { return true; }
        }

        public bool IsSynchronized
        {
            get { return false; }
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return this.GetEnumerator();
        }

        #endregion

        #region Protected

        protected virtual void OnItemsNeeded(ItemsNeededEventArgs e)
        {
            if (this.ItemsNeeded != null)
                this.ItemsNeeded(this, e);
        }

        protected virtual void OnCollectionChanged(NotifyCollectionChangedEventArgs e)
        {
            if (this.CollectionChanged != null)
                this.CollectionChanged(this, e);
        }

        #endregion

        private void AskItems()
        {
            int index = this.loadedList.Count;

            // Controllo che non abbia già chiesto di caricare da questo indice
            if (!this.askedStartIndex.Contains(index))
            {
                this.askedStartIndex.Add(index);

                // Metto un item che indica il caricamento
                this.Add(loadItem);

                this.OnItemsNeeded(new ItemsNeededEventArgs(index));
            }
        }

        public void Add(T value)
        {
            this.loadedList.Add(value);

            NotifyCollectionChangedEventArgs args = new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Add, value, this.loadedList.Count - 1);
            this.OnCollectionChanged(args);
        }

        public void AddRange(IEnumerable<T> items)
        {
            // Rimuovo l'elemento di caricamento, se presente
            RemoveLoadItem();

            foreach (T item in items)
            {
                this.Add(item);
            }
        }

        private void RemoveLoadItem()
        {
            int lastIndex = this.loadedList.Count - 1;
            if (lastIndex >= 0 && this.loadedList[lastIndex].Equals(this.loadItem))
            {
                T item = this.loadedList[lastIndex];
                this.loadedList.RemoveAt(lastIndex);

                NotifyCollectionChangedEventArgs args = new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Remove, item, lastIndex);
                this.OnCollectionChanged(args);
            }
        }

        public bool Contains(T value)
        {
            return this.loadedList.Contains(value);
        }

        public int IndexOf(T value)
        {
            return this.loadedList.IndexOf(value);
        }

        public T this[int index]
        {
            get
            {
                // Se sono all'ultimo alemento e non è l'elemento di caricamento, ne chiedo degli altri
                if (index == this.loadedList.Count - 1 && !this.loadedList[index].Equals(this.loadItem))
                    this.AskItems();
                return this.loadedList[index];
            }
            set
            {
                throw new NotSupportedException();
            }
        }

        public int Count
        {
            get
            {
                // Mi assicuro che abbia chiesto almeno il caricamento
                // iniziale della lista
                if (!askedStartIndex.Contains(0))
                    AskItems();
                return loadedList.Count;
            }
        }

        public object SyncRoot
        {
            get { return syncObject; }
        }

        public IEnumerator<T> GetEnumerator()
        {
            int x = 0;
            foreach (T item in this.loadedList)
            {
                yield return item;

                // Se sono all'ultimo alemento, ne chiedo degli altri
                x++;
                if (x == this.loadedList.Count && item.Equals(this.loadItem))
                    this.AskItems();
            }
        }

    }

    public class ItemsNeededEventArgs : EventArgs
    {

        public ItemsNeededEventArgs(int startIndex)
        {
            this.StartIndex = startIndex;
        }

        public int StartIndex
        {
            get;
            private set;
        }
    }
}
