﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net;
using System.ComponentModel;
using System.IO;

namespace ASPItalia.Books.Chapter13.Server
{
    /// <summary>
    /// Object allowing to send push notifications
    /// </summary>
    public class PushNotificationSender
    {

        public event EventHandler<SendCompletedEventArgs> SendCompleted;

        /// <summary>
        /// Initializes a new instance of the <see cref="PushNotificationSender"/> class.
        /// </summary>
        public PushNotificationSender()
        {

        }

#if !SILVERLIGHT
        /// <summary>
        /// Sends the notification to specified channel URI.
        /// </summary>
        /// <param name="channelUri">The channel URI.</param>
        /// <param name="notification">The notification.</param>
        /// <exception cref="System.Net.WebException" />
        public PushNotificationResponse Send(Uri channelUri, PushNotification notification)
        {
            if (channelUri == null)
                throw new ArgumentNullException("channelUri");
            if (notification == null)
                throw new ArgumentNullException("notification");

            WebClient client = new WebClient();
            client.Encoding = Encoding.UTF8;
            client.Headers[HttpRequestHeader.ContentType] = "text/xml";
            notification.PrepareHeader(client.Headers);

            byte[] data = notification.GetPostData();
            try
            {
                client.UploadData(channelUri, "POST", data);

                WebHeaderCollection headers = client.ResponseHeaders;

                return ParseHeaders(headers);
            }
            catch (WebException we)
            {
                HttpWebResponse hwr = (HttpWebResponse)we.Response;
                switch (hwr.StatusCode)
                {
                    case HttpStatusCode.NotFound:
                        return ParseHeaders(hwr.Headers);
                    case HttpStatusCode.NotAcceptable:
                        return ParseHeaders(hwr.Headers);
                    case HttpStatusCode.PreconditionFailed:
                        return ParseHeaders(hwr.Headers);
                    default:
                        throw;
                }
            }

        }
#endif

        /// <summary>
        /// Sends the notification to specified channel URI.
        /// </summary>
        /// <param name="channelUri">The channel URI.</param>
        /// <param name="notification">The notification.</param>
        /// <exception cref="System.Net.WebException" />
        public void SendAsync(Uri channelUri, PushNotification notification)
        {
            if (channelUri == null)
                throw new ArgumentNullException("channelUri");
            if (notification == null)
                throw new ArgumentNullException("notification");

            byte[] data = notification.GetPostData();
            RequestState state = new RequestState
            {
                Data = notification.GetPostData(),
                Operation = AsyncOperationManager.CreateOperation(null),
                Request = (HttpWebRequest)HttpWebRequest.Create(channelUri),
            };
            state.Request.Method = "POST";
            state.Request.Headers[HttpRequestHeader.ContentEncoding] = "utf8";
            state.Request.ContentType = "text/xml";
            notification.PrepareHeader(state.Request.Headers);
            state.Request.BeginGetRequestStream(OnBeginRequestStream, state);
        }

        private void OnBeginRequestStream(IAsyncResult ar)
        {
            RequestState state = (RequestState)ar.AsyncState;
            try
            {
                using (Stream stream = state.Request.EndGetRequestStream(ar))
                {
                    stream.Write(state.Data, 0, state.Data.Length);
                }

                state.Request.BeginGetResponse(OnBeginGetResponse, state);
            }
            catch (Exception ex)
            {
                state.Operation.PostOperationCompleted(st => OnSendCompleted(new SendCompletedEventArgs(ex, false)), null);
            }
        }

        private void OnBeginGetResponse(IAsyncResult ar)
        {
            RequestState state = (RequestState)ar.AsyncState;
            PushNotificationResponse r = null;
            Exception ex = null;
            try
            {
                HttpWebResponse response = (HttpWebResponse)state.Request.EndGetResponse(ar);

                r = ParseHeaders(response.Headers);
            }
            catch (WebException we)
            {
                HttpWebResponse hwr = (HttpWebResponse)we.Response;
                switch (hwr.StatusCode)
                {
                    case HttpStatusCode.NotFound:
                        r = ParseHeaders(hwr.Headers);
                        break;
                    case HttpStatusCode.NotAcceptable:
                        r = ParseHeaders(hwr.Headers);
                        break;
                    case HttpStatusCode.PreconditionFailed:
                        r = ParseHeaders(hwr.Headers);
                        break;
                    default:
                        ex = we;
                        break;
                }
            }
            catch (Exception e)
            {
                ex = e;
            }

            SendCompletedEventArgs args;
            if (r != null)
                args = new SendCompletedEventArgs(r);
            else
                args = new SendCompletedEventArgs(ex, false);

            state.Operation.PostOperationCompleted(st => OnSendCompleted(args), null);
        }

        protected class RequestState
        {
            internal AsyncOperation Operation;
            internal HttpWebRequest Request;
            internal byte[] Data;
        }

        protected virtual void OnSendCompleted(SendCompletedEventArgs e)
        {
            if (SendCompleted != null)
                SendCompleted(this, e);
        }

        private static PushNotificationResponse ParseHeaders(WebHeaderCollection headers)
        {
            Guid messageId = Guid.Empty;
            NotificationStatus notificationStatus;
            DeviceConnectionStatus deviceConnectionStatus;
            SubscriptionStatus subscriptionStatus;

            GuidExtensions.TryParse(headers["X-MessageID"], out messageId);
            EnumExtensions.TryParse<NotificationStatus>(headers["X-NotificationStatus"], out notificationStatus);
            EnumExtensions.TryParse<DeviceConnectionStatus>(headers["X-DeviceConnectionStatus"], out deviceConnectionStatus);
            EnumExtensions.TryParse<SubscriptionStatus>(headers["X-SubscriptionStatus"], out subscriptionStatus);

            return new PushNotificationResponse(messageId, notificationStatus, deviceConnectionStatus, subscriptionStatus);
        }


    }

    public class SendCompletedEventArgs : AsyncCompletedEventArgs
    {
        public SendCompletedEventArgs(PushNotificationResponse result)
            : base(null, false, null)
        {
            this.Result = result;
        }

        public SendCompletedEventArgs(Exception e, bool cancelled)
            : base(e, cancelled, null)
        {

        }

        public PushNotificationResponse Result
        {
            get;
            private set;
        }
    }
}
