﻿Imports System.Collections.Generic
Imports System.Linq
Imports System.Text
Imports System.Net
Imports System.ComponentModel
Imports System.IO

''' <summary>
''' Object allowing to send push notifications
''' </summary>
Public Class PushNotificationSender

    Public Event SendCompleted As EventHandler(Of SendCompletedEventArgs)

    ''' <summary>
    ''' Initializes a new instance of the <see cref="PushNotificationSender"/> class.
    ''' </summary>

    Public Sub New()
    End Sub

#If Not SILVERLIGHT Then
    ''' <summary>
    ''' Sends the notification to specified channel URI.
    ''' </summary>
    ''' <param name="channelUri">The channel URI.</param>
    ''' <param name="notification">The notification.</param>
    ''' <exception cref="System.Net.WebException" />
    Public Function Send(ByVal channelUri As Uri, ByVal notification As PushNotification) As PushNotificationResponse
        If channelUri Is Nothing Then
            Throw New ArgumentNullException("channelUri")
        End If
        If notification Is Nothing Then
            Throw New ArgumentNullException("notification")
        End If

        Dim client As New WebClient()
        client.Encoding = Encoding.UTF8
        client.Headers(HttpRequestHeader.ContentType) = "text/xml"
        notification.PrepareHeader(client.Headers)

        Dim data As Byte() = notification.GetPostData()
        Try
            client.UploadData(channelUri, "POST", data)

            Dim headers As WebHeaderCollection = client.ResponseHeaders

            Return ParseHeaders(headers)
        Catch we As WebException
            Dim hwr As HttpWebResponse = DirectCast(we.Response, HttpWebResponse)
            Select Case hwr.StatusCode
                Case HttpStatusCode.NotFound
                    Return ParseHeaders(hwr.Headers)
                Case HttpStatusCode.NotAcceptable
                    Return ParseHeaders(hwr.Headers)
                Case HttpStatusCode.PreconditionFailed
                    Return ParseHeaders(hwr.Headers)
                Case Else
                    Throw
            End Select
        End Try

    End Function
#End If

    ''' <summary>
    ''' Sends the notification to specified channel URI.
    ''' </summary>
    ''' <param name="channelUri">The channel URI.</param>
    ''' <param name="notification">The notification.</param>
    ''' <exception cref="System.Net.WebException" />
    Public Sub SendAsync(ByVal channelUri As Uri, ByVal notification As PushNotification)
        If channelUri Is Nothing Then
            Throw New ArgumentNullException("channelUri")
        End If
        If notification Is Nothing Then
            Throw New ArgumentNullException("notification")
        End If

        Dim data As Byte() = notification.GetPostData()
        Dim state As New RequestState() With { _
         .Data = notification.GetPostData(), _
         .Operation = AsyncOperationManager.CreateOperation(Nothing), _
         .Request = DirectCast(HttpWebRequest.Create(channelUri), HttpWebRequest) _
        }
        state.Request.Method = "POST"
        state.Request.Headers(HttpRequestHeader.ContentEncoding) = "utf8"
        state.Request.ContentType = "text/xml"
        notification.PrepareHeader(state.Request.Headers)
        state.Request.BeginGetRequestStream(AddressOf OnBeginRequestStream, state)
    End Sub

    Private Sub OnBeginRequestStream(ByVal ar As IAsyncResult)
        Dim state As RequestState = DirectCast(ar.AsyncState, RequestState)
        Try
            Using stream As Stream = state.Request.EndGetRequestStream(ar)
                stream.Write(state.Data, 0, state.Data.Length)
            End Using

            state.Request.BeginGetResponse(AddressOf OnBeginGetResponse, state)
        Catch ex As Exception
            state.Operation.PostOperationCompleted(Sub(st) OnSendCompleted(New SendCompletedEventArgs(ex, False)), Nothing)
        End Try
    End Sub

    Private Sub OnBeginGetResponse(ByVal ar As IAsyncResult)
        Dim state As RequestState = DirectCast(ar.AsyncState, RequestState)
        Dim r As PushNotificationResponse = Nothing
        Dim ex As Exception = Nothing
        Try
            Dim response As HttpWebResponse = DirectCast(state.Request.EndGetResponse(ar), HttpWebResponse)

            r = ParseHeaders(response.Headers)
        Catch we As WebException
            Dim hwr As HttpWebResponse = DirectCast(we.Response, HttpWebResponse)
            Select Case hwr.StatusCode
                Case HttpStatusCode.NotFound
                    r = ParseHeaders(hwr.Headers)
                    Exit Select
                Case HttpStatusCode.NotAcceptable
                    r = ParseHeaders(hwr.Headers)
                    Exit Select
                Case HttpStatusCode.PreconditionFailed
                    r = ParseHeaders(hwr.Headers)
                    Exit Select
                Case Else
                    ex = we
                    Exit Select
            End Select
        Catch e As Exception
            ex = e
        End Try

        Dim args As SendCompletedEventArgs
        If r IsNot Nothing Then
            args = New SendCompletedEventArgs(r)
        Else
            args = New SendCompletedEventArgs(ex, False)
        End If

        state.Operation.PostOperationCompleted(Sub(st) OnSendCompleted(args), Nothing)
    End Sub

    Protected Class RequestState
        Friend Operation As AsyncOperation
        Friend Request As HttpWebRequest
        Friend Data As Byte()
    End Class

    Protected Overridable Sub OnSendCompleted(ByVal e As SendCompletedEventArgs)
        RaiseEvent SendCompleted(Me, e)
    End Sub

    Private Shared Function ParseHeaders(ByVal headers As WebHeaderCollection) As PushNotificationResponse
        Dim messageId As Guid = Guid.Empty
        Dim notificationStatus As NotificationStatus
        Dim deviceConnectionStatus As DeviceConnectionStatus
        Dim subscriptionStatus As SubscriptionStatus

        GuidExtensions.TryParse(headers("X-MessageID"), messageId)
        EnumExtensions.TryParse(Of NotificationStatus)(headers("X-NotificationStatus"), notificationStatus)
        EnumExtensions.TryParse(Of DeviceConnectionStatus)(headers("X-DeviceConnectionStatus"), deviceConnectionStatus)
        EnumExtensions.TryParse(Of SubscriptionStatus)(headers("X-SubscriptionStatus"), subscriptionStatus)

        Return New PushNotificationResponse(messageId, notificationStatus, deviceConnectionStatus, subscriptionStatus)
    End Function


End Class

Public Class SendCompletedEventArgs
    Inherits AsyncCompletedEventArgs
    Public Sub New(ByVal result As PushNotificationResponse)
        MyBase.New(Nothing, False, Nothing)
        Me.Result = result
    End Sub

    Public Sub New(ByVal e As Exception, ByVal cancelled As Boolean)

        MyBase.New(e, cancelled, Nothing)
    End Sub

    Public Property Result() As PushNotificationResponse
        Get
            Return m_Result
        End Get
        Private Set(ByVal value As PushNotificationResponse)
            m_Result = value
        End Set
    End Property
    Private m_Result As PushNotificationResponse
End Class